import json
import pytest

from unittest.mock import mock_open, patch

from must_triage.inspectors.ocs import OCS


class TestOCS:
    def teardown(self):
        patch.stopall()

    @pytest.mark.parametrize(
        "obj,expected",
        [
            (
                dict(status='HEALTH_OK'),
                list(),
            ),
            (
                dict(status='something else'),
                [dict(status='something else')],
            ),
        ]
    )
    def test_unhealthy(self, obj, expected):
        result = OCS.unhealthy(obj)
        assert (obj['status'] == 'HEALTH_OK') != (obj in result)

    @pytest.mark.parametrize(
        "path,obj",
        [
            (
                '/fake_path',
                None
            ),
            (
                '/fake_path',
                json.decoder.JSONDecodeError('', '', 0)
            ),
            (
                '/fake/health_detail',
                dict()
            ),
        ]
    )
    def test_inspect_json(self, path, obj):
        expected = None
        p_stat = patch('must_triage.inspectors.ocs.os.stat')
        m_stat = p_stat.start()
        if obj is None:
            m_stat.return_value.st_size = 0
            expected = ["File is empty"]
        else:
            m_stat.return_value.st_size = 1
        p_load = patch('must_triage.inspectors.ocs.json.load')
        m_load = p_load.start()
        if isinstance(obj, dict) or obj is None:
            m_load.return_value = obj
        else:
            m_load.side_effect = obj
            expected = ["Failed to parse JSON content"]
        p_unhealthy = patch('must_triage.inspectors.ocs.OCS.unhealthy')
        m_unhealthy = p_unhealthy.start()
        m_open = mock_open()
        with patch('must_triage.inspectors.ocs.open', m_open):
            result = OCS.inspect_json(path)
        if expected is None:
            assert m_unhealthy.call_count == 1
        else:
            assert expected == result[path]

    @pytest.mark.parametrize(
        "lines,panics",
        [
            (
                [
                    'we Observed a panic: here it is',
                ],
                1,
            ),
            (
                [
                    'no panic',
                ],
                0,
            ),
            (
                [
                    'we Observed a panic: here it is',
                    'no panic',
                    'Observed a panic:',
                ],
                2,
            ),
        ]
    )
    def test_inspect_log(self, lines, panics):
        path = '/fake_path'
        m_open = mock_open()
        with patch('must_triage.inspectors.ocs.open', m_open):
            m_open.return_value.readlines.return_value = lines
            result = OCS.inspect_log(path)
        assert len(result[path]) == panics



import argparse
import asyncio
import json
import os
import sys
import yaml

from must_triage import formatters
from must_triage import inspectors


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument('path', type=dir_or_file)
    parser.add_argument('-q', '--quiet', action='store_true')
    parser.add_argument(
            '-o', '--out', choices=['json', 'yaml'], default='json')
    return parser.parse_args(args)


def dir_or_file(path):
    result = os.path.expanduser(path)
    if os.path.isdir(result) or os.path.isfile(result):
        return result
    raise argparse.ArgumentTypeError(f"{path} is not a valid path!")


def main():
    args = parse_args(sys.argv[1:])
    interests = dict()
    for inspector_cls in inspectors.all():
        inspector = inspector_cls(root=args.path, progress=not args.quiet)
        inspector.gather()
        inspectors.merge_interests(interests, asyncio.run(inspector.inspect()))

    if args.out == 'json':
        print(json.dumps(
            interests, indent=2, default=formatters.json_serialize))
    elif args.out == 'yaml':
        print(yaml.safe_dump(interests, default_flow_style=False))



import os


def has_ext(path, ext):
    if not isinstance(ext, list):
        ext = [ext]
    return path.split('.')[-1] in ext


def find(path, callback):
    if os.path.isfile(path) and callback(path):
        return [path]
    results = list()
    for root, dirs, files in os.walk(path):
        matches = list(filter(callback, files))
        if matches:
            matches = [os.path.join(root, s) for s in matches]
            results.extend(matches)
    return results



from tqdm import tqdm


class ProgressBar(tqdm):
    "A simple subclass of tqdm with our configuration items set"
    def __init__(self, *args, **kwargs):
        kwargs.update(dict(
            miniters=None,
            mininterval=0.1,
            maxinterval=1,
            smoothing=0,
            bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}]',
        ))
        super(ProgressBar, self).__init__(*args, **kwargs)

    @property
    def format_dict(self):
        d = super(ProgressBar, self).format_dict
        total_time = d["elapsed"] * (d["total"] or 0) / max(d["n"], 1)
        d.update(total_time=self.format_interval(total_time) + " in total")
        return d



import logging
import pytest

from concurrent.futures import ProcessPoolExecutor
from unittest.mock import patch

from must_triage.inspectors.base import Inspector


@pytest.fixture
def inspector(request):
    obj = Inspector(root='/tmp/must_triage_tests', progress=False)
    return obj


class TestInspector:
    def teardown(self):
        patch.stopall()

    def test_init(self, inspector):
        assert inspector.root == '/tmp/must_triage_tests'
        assert inspector.progress is False
        assert isinstance(inspector.log, logging.Logger)

    def test_gather(self, inspector):
        inspector.gather_types = dict(
            jpg=dict(
                match=None,
                description='desc',
            )
        )
        with patch('must_triage.inspectors.base.fs.find') as m_find:
            m_find.return_value = 'path'
            inspector.gather()
        assert m_find.call_count == len(inspector.gather_types.keys())
        assert len(inspector.gathered.keys()) == m_find.call_count

    @pytest.mark.asyncio
    async def test_inspect(self, inspector):
        inspector.gather_types = dict(
            jpg=dict(),
            png=dict(),
        )
        inspector.gathered = dict(
            jpg=['1', '2'],
            gif=['3', '4'],
            png=['5', '6'],
        )
        inspector.inspect_jpg = True
        inspector.inspect_png = True
        p_helper = patch(
            'must_triage.inspectors.base.Inspector._inspect_helper')
        m_helper = p_helper.start()
        m_helper.return_value = dict()
        await inspector.inspect()
        assert m_helper.call_count == 2  # once each for jpg, png

    @staticmethod
    def _dummy_inspector(path):
        return {path: [f"inspected {path}"]}

    @pytest.mark.asyncio
    async def test_inspect_helper(self, inspector):
        paths = ['/foo', '/bar']

        with ProcessPoolExecutor() as executor:
            result = await inspector._inspect_helper(
                executor, paths, self._dummy_inspector, "description")

        for key, value in result.items():
            assert len(value) == 1
            assert value[0] == f"inspected {key}"

    @pytest.mark.asyncio
    async def test_inspect_helper_noop(self, inspector):
        result = await inspector._inspect_helper(
            None, [], None, None)
        assert result == dict()

    def test_inspector_wrapper_exc(self, inspector):

        def func(*args):
            raise Exception

        result = inspector._inspector_wrapper(func, '')
        assert result == dict()



import pytest

from must_triage import progress


@pytest.fixture
def obj(request):
    return progress.ProgressBar()


class TestProgressBar:
    def test_init(self, obj):
        assert obj.mininterval == 0.1
        assert obj.maxinterval == 1
        assert obj.smoothing == 0
        assert obj.bar_format == \
            '{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}]'

    def test_format_dict(self, obj):
        assert 'total_time' in obj.format_dict



import datetime
import pytest

from must_triage import formatters


@pytest.mark.parametrize(
    "obj,expected",
    [
        (
            datetime.datetime.utcfromtimestamp(1600469633.228982),
            "2020-09-18 22:53:53.228982"
        ),
        (
            True,
            None
        ),
    ]
)
def test_json_serializer(obj, expected):
    assert formatters.json_serialize(obj) == expected



import logging

from concurrent.futures import ProcessPoolExecutor
from functools import partial

import must_triage.fs as fs
import must_triage.inspectors as inspectors
from must_triage.progress import ProgressBar


class Inspector:
    def __init__(self, root, progress=False):
        self.root = root
        self.progress = progress
        self.log = logging.getLogger(self.__class__.__name__)

    def gather(self):
        self.gathered = dict()
        for type_, obj in self.gather_types.items():
            # this assert is mainly to help diagnose unexpected failures
            assert 'match' in obj
            self.gathered[type_] = fs.find(self.root, obj['match'])

    async def inspect(self):
        self.interests = dict()
        with ProcessPoolExecutor() as executor:
            for type_, paths in self.gathered.items():
                func = getattr(self, f"inspect_{type_}", None)
                if not func:
                    continue
                description = self.gather_types[type_].get(
                    'description', f"Inspecting {type_}")
                interests = await self._inspect_helper(
                    executor,
                    paths,
                    func,
                    description,
                )
                inspectors.merge_interests(self.interests, interests)
        return self.interests

    def _inspector_wrapper(self, func, path):
        try:
            return func(path)
        except Exception:
            self.log.exception(
                f"Unhandled exception while processing  {path}")
            return dict()

    async def _inspect_helper(self, executor, paths, func, description):
        if not paths:
            return dict()
        interests = dict()

        wrapped = partial(self._inspector_wrapper, func)

        results = list(ProgressBar(
            executor.map(wrapped, paths),
            total=len(paths),
            desc=description,
            disable=not self.progress,
        ))
        for result in results:
            interests.update(result)
        return interests



import pyfakefs
import pytest

from unittest.mock import patch

import must_triage.fs


@pytest.mark.parametrize(
    "path,ext,expected",
    [
        (
            '/a/jpeg', 'jpeg', False
        ),
        (
            '/a.jpeg', 'jpeg', True
        ),
        (
            'a.yml', ['yaml', 'yml'], True
        ),
        (
            'a.yaml', '.json', False
        ),
    ]
)
def test_has_ext(path, ext, expected):
    assert must_triage.fs.has_ext(path, ext) is expected


@pytest.mark.parametrize(
    "files,path,callback,expected",
    [
        (
            [
                dict(path='/a/b/c/d.yaml'),
                dict(path='/a/b/c/d.json'),
                dict(path='/a/b/c/d/e.yaml'),
                dict(path='/1/2/3/4.yaml'),
            ],
            '/a',
            lambda p: p.endswith('.yaml'),
            [
                '/a/b/c/d.yaml',
                '/a/b/c/d/e.yaml',
            ]
        ),
        (
            [
                dict(path='/a/b.log'),
            ],
            '/a/b.log',
            lambda p: p.endswith('log'),
            [
                '/a/b.log',
            ]
        ),
        (
            [
                dict(path='/1/2/3.txt'),
                dict(path='/4/5/6.tar.gz'),
                dict(path='/7/8/9.so'),
            ],
            '/',
            lambda p: False,
            []
        ),
    ]
)
def test_fs(fs, files, path, callback, expected):
    for f in files:
        fs.create_file(
            f['path'],
            contents=f.get('contents', ''),
        )
    result = must_triage.fs.find(path, callback)
    assert result == expected



import yaml

import must_triage.fs as fs

from must_triage.inspectors.base import Inspector


class OCP(Inspector):
    gather_types = dict(
        yaml=dict(
            match=lambda p: fs.has_ext(p, ['yaml', 'yml']),
            description="Reading OCP YAML files",
        ),
    )

    @staticmethod
    def inspect_yaml(path):
        result = {path: list()}
        with open(path) as fd:
            try:
                obj = yaml.safe_load(fd)
            except (yaml.scanner.ScannerError, yaml.parser.ParserError):
                result[path].append("Failed to parse YAML content")
                return result
        if 'kind' in obj:
            result[path].extend(OCP.operator_success(obj))
            pods = list()
            if obj['kind'].lower() == 'pod':
                pods.append(obj)
            elif obj['kind'].lower() == 'podlist':
                pods.extend(filter(
                    lambda o: o['kind'].lower() == 'pod',
                    obj['items']
                ))
            result[path].extend(map(OCP.pod_ready, pods))
        if 'error' in obj:
            result[path].append(obj)
        return result

    @staticmethod
    def operator_success(obj):
        result = list()
        if obj['kind'] != 'ClusterServiceVersion':
            return result
        status = obj['status']
        if status['phase'].lower() == 'succeeded':
            return result
        result.append(
            f"Operator phase is '{status['phase']}', "
            "not 'Succeeded' as expected"
        )
        return result

    @staticmethod
    def pod_ready(obj):
        result = list()
        status = obj['status']
        try:
            container_statuses = status['containerStatuses']
        except KeyError:
            return [dict(
                pod_name=obj['metadata']['name'],
                status=status,
                )]
        not_ready = filter(lambda cs: cs['ready'] is False, container_statuses)
        for cs in not_ready:
            ts = cs['state'].get('terminated', dict())
            if (ts.get('reason', '').lower() == 'completed' and
                    ts.get('exitCode') == 0):
                continue
            result.append(
                f"Container '{cs['name']}' in "
                f"pod '{obj['metadata']['name']}' is not ready")
        return result



import argparse
import pytest

from unittest.mock import patch

from must_triage import cmd


def test_parse_args():
    args = ['.']
    parsed = cmd.parse_args(args)
    assert parsed.quiet is False
    assert parsed.out == 'json'
    assert parsed.path == '.'


def test_parse_args_bad_path():
    args = ['/fake_path_here']
    with pytest.raises((argparse.ArgumentError, SystemExit)):
        cmd.parse_args(args)


class TestMain:
    def teardown(self):
        patch.stopall()

    @pytest.mark.parametrize(
        "args",
        [
            ['must-triage', '.'],
            ['must-triage', '-o', 'yaml', '.'],
        ]
    )
    def test_main(self, args):
        p_argv = patch(
            'must_triage.cmd.sys.argv', args)
        p_argv.start()
        p_ocp = patch('must_triage.cmd.inspectors.ocp.OCP', autospec=True)
        m_ocp = p_ocp.start()
        m_ocp.return_value.inspect.return_value = dict()
        p_ocs = patch('must_triage.cmd.inspectors.ocs.OCS', autospec=True)
        m_ocs = p_ocs.start()
        m_ocs.return_value.inspect.return_value = dict()
        p_all = patch('must_triage.cmd.inspectors.all')
        m_all = p_all.start()
        m_all.return_value = [m_ocp, m_ocs]
        p_merge = patch('must_triage.cmd.inspectors.merge_interests')
        m_merge = p_merge.start()
        p_dumps = patch('must_triage.cmd.json.dumps')
        m_dumps = p_dumps.start()
        p_safe_dump = patch('must_triage.cmd.yaml.safe_dump')
        m_safe_dump = p_safe_dump.start()
        cmd.main()
        assert m_all.call_count == 1
        assert m_ocp.return_value.gather.call_count == 1
        assert m_ocp.return_value.inspect.await_count == 1
        assert m_ocs.return_value.gather.call_count == 1
        assert m_ocs.return_value.inspect.await_count == 1
        assert m_merge.call_count == 2
        if 'yaml' in args:
            assert m_safe_dump.call_count == 1
        else:
            assert m_dumps.call_count == 1



import json
import os
import re

import must_triage.fs as fs

from must_triage.inspectors.base import Inspector


class OCS(Inspector):
    gather_types = dict(
        json=dict(
            match=lambda p: re.match('.*--format_json(-pretty)?$', p),
            description="Reading OCS JSON files",
        ),
        log=dict(
            match=lambda p: fs.has_ext(p, ['log']),
            description="Reading OCS log files",
        ),
    )

    @staticmethod
    def inspect_json(path):
        result = {path: list()}
        if os.stat(path).st_size == 0:
            result[path].append("File is empty")
            return result
        with open(path) as fd:
            try:
                obj = json.load(fd)
            except json.decoder.JSONDecodeError:
                result[path].append("Failed to parse JSON content")
                return result
        if 'health_detail' in path:
            result[path].extend(OCS.unhealthy(obj))
        return result

    @staticmethod
    def unhealthy(obj):
        result = list()
        if obj['status'] != 'HEALTH_OK':
            result.append(obj)
        return result

    @staticmethod
    def inspect_log(path):
        result = {path: list()}
        with open(path) as log:
            for line in log.readlines():
                line = line.strip()
                if OCS.panicked(line):
                    result[path].append(line)
        return result

    @staticmethod
    def panicked(line):
        if re.match('.*Observed a panic:.*', line):
            return True
        return False



import pytest
import yaml

from unittest.mock import mock_open, patch

from must_triage.inspectors.ocp import OCP


class TestOCP:
    def teardown(self):
        patch.stopall()

    @pytest.mark.parametrize(
        "obj",
        [
            (
                yaml.scanner.ScannerError
            ),
            (
                dict(kind='pod')
            ),
            (
                dict(
                    kind='podList',
                    items=[
                        dict(kind='pod'),
                        dict(kind='pod'),
                    ],
                )
            ),
            (
                dict(error='something')
            ),
        ]
    )
    def test_inspect_yaml(self, obj):
        path = '/fake_path'
        p_safe_load = patch('must_triage.inspectors.ocp.yaml.safe_load')
        m_safe_load = p_safe_load.start()
        p_operator_success = patch(
            'must_triage.inspectors.ocp.OCP.operator_success')
        m_operator_success = p_operator_success.start()
        p_pod_ready = patch('must_triage.inspectors.ocp.OCP.pod_ready')
        m_pod_ready = p_pod_ready.start()
        m_open = mock_open()
        with patch('must_triage.inspectors.ocp.open', m_open):
            if isinstance(obj, dict):
                m_safe_load.return_value = obj
            else:
                m_safe_load.side_effect = obj
            result = OCP.inspect_yaml(path)
        if not isinstance(obj, dict):
            assert result[path] == ["Failed to parse YAML content"]
            return
        assert len(m_operator_success.call_args_list) == \
            len(list(filter(lambda o: 'kind' in o, obj)))

        def pods_count(obj):
            kind = obj.get('kind', '').lower()
            if kind == 'pod':
                return 1
            if kind == 'podlist':
                return len(
                    [o for o in obj['items'] if o['kind'].lower() == 'pod']
                )
            return 0

        assert len(m_pod_ready.call_args_list) == pods_count(obj)

        if 'error' in obj:
            assert obj in result[path]

    @pytest.mark.parametrize(
        "obj,expected",
        [
            (
                dict(
                    kind='ClusterServiceVersion',
                    status=dict(phase='succeeded'),
                ),
                list()
            ),
            (
                dict(
                    kind='ClusterServiceVersion',
                    status=dict(phase='other'),
                ),
                ["Operator phase is 'other', not 'Succeeded' as expected"]

            ),
            (
                dict(kind='other'),
                list()
            )
        ]
    )
    def test_operator_success(self, obj, expected):
        result = OCP.operator_success(obj)
        assert result == expected

    @pytest.mark.parametrize(
        "obj,expected",
        [
            (
                dict(
                    metadata=dict(name='test_pod'),
                    status=dict(containerStatuses=[
                        dict(
                            name='test_container',
                            ready=False,
                            state=dict(
                                terminated=dict(
                                    reason='garbage',
                                ),
                            ),
                        ),
                    ]),
                ),
                ["Container 'test_container' in pod 'test_pod' is not ready"],
            ),
            (
                dict(
                    metadata=dict(name='test_pod'),
                    status=dict(containerStatuses=[
                        dict(
                            name='test_container',
                            ready=False,
                            state=dict(
                                terminated=dict(
                                    reason='completed',
                                    exitCode=0,
                                ),
                            ),
                        ),
                    ]),
                ),
                list(),
            ),
            (
                dict(
                    metadata=dict(name='test_pod'),
                    status=dict(),
                ),
                [dict(
                    pod_name='test_pod',
                    status=dict(),
                )],
            ),
        ]
    )
    def test_pod_ready(self, obj, expected):
        result = OCP.pod_ready(obj)
        assert result == expected



import pytest

import must_triage.inspectors as inspectors


class TestMergeInterests:
    @pytest.mark.parametrize(
        "existing,new,expected",
        [
            (
                dict(foo=['foo']),
                dict(bar=['bar']),
                dict(foo=['foo'], bar=['bar']),
            ),
            (
                dict(foo=['1']),
                dict(foo=['2']),
                dict(foo=['1', '2']),
            ),
            (
                dict(foo=['foo']),
                dict(bar=[]),
                dict(foo=['foo']),
            )
        ]
    )
    def test_merge(self, existing, new, expected):
        assert inspectors.merge_interests(existing, new) == expected


def test_all():
    inspector_names = list(map(lambda i: i.__name__, inspectors.all()))
    assert inspector_names == ['OCP', 'OCS']



from . import ocp, ocs


def merge_interests(existing, new):
    for key, value in new.items():
        if value == []:
            continue
        if type(value) == list:
            value = filter(lambda v: v, value)
        if key not in existing:
            existing[key] = list()
        existing[key].extend(value)
    return existing


def all():
    return [ocp.OCP, ocs.OCS]
